###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2021 RWTH Aachen University
# Git: https://git.rwth-aachen.de/coscine/docs/public/coscine-python-sdk
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
#
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://www.coscine.de for more information.
#
# Please note that this python module resembles a community effort
# and not an official service that RWTH Aachen provides support for.
# It has been initially brough to life by the team at Coscine but has since
# been turned over to the community around Coscine.
###############################################################################

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
	from .resource import Resource
from .exceptions import *
from .form import InputForm

###############################################################################

class MetadataForm(InputForm):

	"""
	Input form for interacting with Metadata.
	"""

	def __init__(self, resource: Resource, data: dict = None) -> None:
		"""
		Initializes the Metadata form.

		Parameters
		-----------
		resource : coscine.Resource
			Coscine resource instance.
		data : dict (optional)
			Data to initialize the MetadataForm with.
			If the form accepts fields "Title" and "Author", data could
			be data = {"Title": "a", "Author": "b"}
		"""

		client = resource.client
		entries = []
		vocabulary = {}
		lang = resource.client.lang
		profile = resource.application_profile(parse = True)
		for element in profile["graph"]:
			flags = MetadataForm.NONE
			if "minCount" in element and element["minCount"] > 0:
				flags |= MetadataForm.REQUIRED
			if "maxCount" in element and element["maxCount"] > 1:
				flags |= MetadataForm.LIST

			if "class" in element:
				flags |= MetadataForm.CONTROLLED
				uri = element["class"]
				instance = client.static.instance(resource.project.id, uri)
				if lang not in instance or len(instance[lang]) == 0:
					lang = "en"
				voc = {}
				for rule in instance[lang]:
					voc[rule["name"]] = rule["value"]
				vocabulary[element["name"][lang]] = voc

			entry = {
				"name": element["name"],
				"path": element["path"],
				"field": element["name"][lang],
				"flags": flags,
				"order": element["order"],
				"datatype": element["datatype"],
				"type": element["type"]
			}
			entries.append(entry)

		# Sort the keys according to their application profile order
		entries = sorted(entries, key = lambda x: x["order"])
		super().__init__("Metadata Form", lang, entries, vocabulary)
		self.profile = profile

		# Fill
		if data:
			for key in data:
				self[key] = data[key]

###############################################################################

	def parse(self, data: dict) -> None:

		"""
		Parses JSON-LD metadata into a Metadata Input Form

		Parameters
		-----------
		data : dict
			JSON-LD metadata (as retrieved from Coscine).
		"""

		if data is None:
			return

		for path in data:
			for entry in self.profile["graph"]:
				if path == entry["path"]:
					key = entry["name"][self._lang]
					value = data[path][0]["value"]
					if self.is_controlled(key):
						voc = self.get_vocabulary(key)
						keys = list(voc.keys())
						values = list(voc.values())
						index = values.index(value)
						value = keys[index]
					self.store[key] = value
					break

###############################################################################

	def generate(self) -> dict:

		"""
		Generates JSON-LD metadata for use in Coscine.

		Returns
		-------
		dict
			JSON-LD formatted metadata
		"""

		metadata = {}
		RDFTYPE = "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"

		# Set application profile type used by metadata
		metadata[RDFTYPE] = [{
			"type": "uri",
			"value": self.profile["id"]
		}]

		# Collect missing required fields
		missing = []

		# Set metadata fields
		for key in self._keys:
			entry = self._keys[key]
			field = entry["field"]
			if key not in self.store:
				if self.is_required(key):
					missing.append(key)
			else:
				path = entry["path"]
				value = self.store[key]
				if entry["flags"] & MetadataForm.LIST:
					if type(value) is not list and type(value) is not tuple:
						raise ValueError("Expected iterable value for key %s" % key)
				if self.is_controlled(key):
					voc = self.get_vocabulary(key)
					value = voc[value]
				metadata[path] = [{
					"value": value,
					"datatype": entry["datatype"],
					"type": entry["type"]
				}]

		# Check missing field list
		if len(missing) > 0:
			raise RequirementError(missing)

		return metadata

###############################################################################

class MetadataPresetForm(InputForm):

	"""
	The MetadataPreset form allows for the configuration of default
	values in an application profile.
	"""

	def __init__(self, resource: Resource, data: dict = None) -> None:
		"""
		Initialized the MetadataPreset form.

		Parameters
		----------
		resource : coscine.Resource
			Coscine resource instance.
		data : dict, default: None, currently unused
			Data to initialize the form with.
			If the form accepts fields "Title" and "Author", data could
			be data = {"Title": "a", "Author": "b"}
		"""

		entries = []
		vocabulary = {}
		lang = resource.client.lang
		applicationProfile = resource.application_profile(parse = True)
		for element in applicationProfile["graph"]:
			flags = MetadataPresetForm.NONE
			if "minCount" in element and element["minCount"] > 0:
				flags |= MetadataPresetForm.REQUIRED
			if "maxCount" in element and element["maxCount"] > 1:
				flags |= MetadataPresetForm.LIST

			if "class" in element:
				flags |= MetadataPresetForm.CONTROLLED
				uri = element["class"]
				instance = resource.client.static.instance(uri)
				if lang not in instance or len(instance[lang]) == 0:
					lang = "en"
				voc = {}
				for rule in instance[lang]:
					voc[rule["name"]] = rule["value"]
				vocabulary[element["name"][lang]] = voc
			
			entry = {
				"name": element["name"],
				"path": element["id"],
				"flags": flags,
				"field": element["name"][lang],
				"order": element["order"],
				"datatype": element["datatype"],
				"type": element["type"],
				"locked": False,
				"enabled": True
			}
			entries.append(entry)

		# Sort the keys according to their application profile order
		entries = sorted(entries, key = lambda x: x["order"])
		super().__init__("Metadata Preset", lang, entries, vocabulary)
		self.profile = applicationProfile

###############################################################################

	def enable(self, key: str, enable: bool = True) -> None:

		"""
		Enables or disables a metadata field.

		Parameters
		-----------
		key : str
			Form field key name.
		lock : bool, default: True
			True to enable, False to disable
		"""

		self._keys[key]["enabled"] = enable

###############################################################################

	def lock(self, key: str, lock: bool = True) -> None:

		"""
		Locks a form fields value.

		Note
		----
		Currently has no effect.

		Parameters
		-----------
		key : str
			form field key name
		lock : bool, default: True
			True to lock, False to unlock
		"""

		self._keys[key]["locked"] = lock

###############################################################################

	def generate(self) -> dict:

		"""
		Generates JSON-LD formatted representation of MetadataPreset data.

		Raises
		-------
		RequirementError
			When one or more required fields have not been set.
		
		Returns
		--------
		dict
			JSON-LD formatted MetadataPreset data.
		"""

		metadata = {}

		# Set metadata fields
		for key in self._keys:
			field = self._keys[key]
			path = field["path"]

			if key not in self.store:
				value = None
				metadata[path] = {
					"https://purl.org/coscine/defaultValue": [],
					"https://purl.org/coscine/invisible": [{
						"type": "literal",
						"value": "0"
					}]
				}
				continue
			else:
				value = self.store[key]
			if self.is_controlled(key):
				voc = self.get_vocabulary(key)
				value = voc[value]
			trigger: str = ["1", "0"][self._keys[key]["enabled"]]
			metadata[path] = {
				"https://purl.org/coscine/defaultValue": [{
					"type": field["type"],
					"value": value
				}],
				"https://purl.org/coscine/invisible": [{
					"type": "literal",
					"value": trigger
				}]
			}

		return metadata

###############################################################################


###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2021 RWTH Aachen University
# Git: https://git.rwth-aachen.de/coscine/docs/public/coscine-python-sdk
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
#
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://www.coscine.de for more information.
#
# Please note that this python module resembles a community effort
# and not an official service that RWTH Aachen provides support for.
# It has been initially brough to life by the team at Coscine but has since
# been turned over to the community around Coscine.
###############################################################################

from collections.abc import MutableMapping
from .exceptions import *
from typing import List
import prettytable

###############################################################################

class InputForm(MutableMapping):

	"""
	The InputForm class provides an overarching input form interface for
	metadata, project or resource editing. It is meant to resemble
	the input form of the web interface as closely as possible.
	The input form is built opon the python dict datatype and can be interacted
	with like with any other dictionary. However some additional functions
	are provided, to make it easier to interact with the form and its contents.
	"""

	# Form flags
	NONE = 0
	REQUIRED = 1
	CONTROLLED = 2
	SET = 4
	FIXED = 8
	LIST = 16
	SPECIAL = 32

###############################################################################

	def __init__(self, name: str, lang: str, entries: list, vocabulary: dict):

		"""
		Each form is intialized with a name, a language present and the actual
		entries and their vocabularies.

		Parameters
		-----------
		name : str
			The name of the form (e.g. 'Resource Form').
		lang : str
			The language of the client ('en' or 'de').
		entries : list
			A list of form entries (list of dict).
		vocabulary : dict
			A mapping of form field key to vocabulary dict.
		"""

		self._name = name
		self._lang = lang
		self._entries = entries
		self._keys = {}
		for entry in entries:
			self._keys[entry["name"][self._lang]] = entry
		self._vocabulary = vocabulary
		self.store = {}

###############################################################################

	def __getitem__(self, key: str) -> dict:
		if key not in self.keys():
			raise KeyError(key)
		elif key in self.store:
			return self.store[key]
		else:
			return None

###############################################################################

	def __setitem__(self, key: str, value: object):
		if key not in self._keys:
			raise KeyError("Key `%s` not expected by form." % key)
		if self._expects_list(key) and (type(value) is not list):
			raise ValueError("Expected a list value for key `%s`." % key)
		if self.is_controlled(key):
			vocabulary = self.get_vocabulary(key)
			if type(value) is list:
				self.store[key] = []
				for val in value:
					if val in vocabulary:
						self.store[key].append(val)
					else:
						raise VocabularyError(val)
			else:
				if value in vocabulary:
					self.store[key] = value
				else:
					raise VocabularyError(value)
		else:
			self.store[key] = value

###############################################################################

	def __delitem__(self, key: str):
		del self.store[key]

###############################################################################

	def __iter__(self):
		return iter(self.store)

###############################################################################

	def __len__(self):
		return len(self.store)

###############################################################################

	def __str__(self) -> str:
		table = prettytable.PrettyTable(("R", "C", "Field", "Value"))
		table.max_width["Value"] = 40
		rows = []
		for key in self.keys():
			flag1 = flag2 = " "
			if self.is_required(key):
				flag1 = "X"
			if self.is_controlled(key):
				flag2 = "X"
			if key in self.store:
				value = self.store[key]
			else:
				value = ""
			rows.append((flag1, flag2, key, value))
		table.add_rows(rows)
		return table.get_string(title=self._name)

###############################################################################

	def __repr__(self) -> str:
		return self.__str__()

###############################################################################

	def _expects_list(self, key: str) -> bool:
		"""
		Returns wether the field referenced by key requires a list.
		"""

		return self._keys[key]["flags"] & self.LIST

###############################################################################

	def is_required(self, key: str) -> bool:

		"""
		Determines wether the given key is a required field.

		Parameters
		-----------
		key : str
			A key contained in the input form (e.g. 'Author').

		Returns
		-------
		True
			If key is a required field.
		False
			If key is not a required field.
		"""

		return self._keys[key]["flags"] & self.REQUIRED

###############################################################################

	def is_controlled(self, key: str) -> bool:

		"""
		Determines wether the given key is a controlled field. Controlled
		fields have a vocabulary. Any value assigned to the key must be
		contained in that vocabulary.

		Parameters
		-----------
		key : str
			A key contained in the input form (e.g. 'Author')

		Returns
		-------
		True
			If key is a controlled field.
		False
			If key is not a controlled field.
		"""

		return self._keys[key]["flags"] & self.CONTROLLED

###############################################################################

	def get_field(self, key: str) -> str:

		"""
		Returns the internal field name of a field contained in the form.

		Parameters
		-----------
		key : str
			A key contained in the input form (e.g. 'Author').

		Returns
		-------
		A string with the name used to represent the field internally in
		Coscine JSON-LD. (e.g. 'doe' instead of 'John Doe')
		"""

		return self._keys[key]["field"]

###############################################################################

	def get_vocabulary(self, key: str) -> dict:

		"""
		Returns the vocabulary of a controlled field. Uncontrolled fields
		result in an exception.

		Parameters
		----------
		key : str
			Any field contained inside of the form

		Returns
		--------
		dict
			The vocabulary of the form field.
		"""

		if self.is_controlled(key):
			return self._vocabulary[self.get_field(key)]
		else:
			msg = "Key [%s] is not controlled by a vocabulary!" % key
			raise CoscineException(msg)

###############################################################################

	def keys(self) -> List[str]:

		"""
		Returns a list of all keys contained within the input form.

		Returns
		-------
		list[str]
		"""

		return self._keys.keys()

###############################################################################

	def reset(self):

		"""
		Resets the input form to default values.
		"""

		self.store.clear()

###############################################################################

	def parse(self, data: dict):
		pass

###############################################################################

	def generate(self):
		pass

###############################################################################

	def entry(self, field: str) -> dict:

		"""
		Returns an entry of the input form. An entry resembles the internal
		representation of a form field.

		Parameters
		----------
		field : str
			The Coscine internal name of the field (see InputForm.get_field(...)).

		Returns
		-------
		dict
			The fields internal entry representation as a dict.
		"""

		for item in self._entries:
			if item["field"] == field:
				return item
		return None

###############################################################################


###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2021 RWTH Aachen University
# Git: https://git.rwth-aachen.de/coscine/docs/public/coscine-python-sdk
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
#
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://www.coscine.de for more information.
#
# Please note that this python module resembles a community effort
# and not an official service that RWTH Aachen provides support for.
# It has been initially brough to life by the team at Coscine but has since
# been turned over to the community around Coscine.
###############################################################################

###############################################################################
"""
Coscine Python SDK

This python3.x package provides an SDK for interfacing with Coscine.
It is maintained and developed by the community and should not be regarded
as an official service of RWTH Aachen.
You can use the package to integrate Coscine in your python applications
and easily:
- create, edit or delete projects and resources
- invite and manage project members
- upload and download files, resources and projects
- manage your metadata
"""
###############################################################################

from .client import Client
from .project import Project, ProjectForm, ProjectMember
from .resource import Resource, ResourceForm
from .object import Object
from .metadata import MetadataForm, MetadataPresetForm
from .exceptions import *